"""
Pocket dimension provides a memory-efficient, dense, random projection of sparse vectors
and then applies this to Term Frequency (TF) and Term Frequency, Inverse Document
Frequency (TF-IDF) data.

Copyright (C) 2022 Matthew Hendrey & Brendan Murphy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
__version__ = "0.1.4"


from pocket_dimension.random_projection import (
    random_projection,
    distributional_johnson_lindenstrauss_optimal_delta,
    random_sparse_vectors,
    JustInTimeRandomProjection,
)
from pocket_dimension.vectorizer import (
    TFVectorizer,
    TFIDFVectorizer,
)



"""
Pocket Dimension provides a memory-efficient, dense, random projection of sparse vectors
and then applies this to Term Frequency (TF) and Term Frequency, Inverse Document
Frequency (TFIDF) data.

Copyright (C) 2022 Matthew Hendrey & Brendan Murphy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from numba import njit, prange, float32, int64, uint64
import numpy as np
from sklearn.random_projection import (
    BaseRandomProjection,
    johnson_lindenstrauss_min_dim,
)
from sklearn.exceptions import DataDimensionalityWarning
from sklearn.utils.validation import check_is_fitted
from scipy.optimize import minimize
from scipy.sparse import csr_matrix
from scipy.stats import beta
from typing import Dict, Iterable, List, Tuple, Type, Union
import warnings


@njit(uint64(uint64))
def splitmix64(index):
    """
    Fast, simple function for taking an integer and returning a random number. Function
    used in Java random number generator.

    Parameters
    ----------
    index : uint64

    Returns
    -------
    uint64
    """
    z = index + uint64(0x9E3779B97F4A7C15)
    z = (z ^ (z >> uint64(30))) * uint64(0xBF58476D1CE4E5B9)
    z = (z ^ (z >> uint64(27))) * uint64(0x94D049BB133111EB)

    return z ^ (z >> uint64(31))


@njit(float32[:, :](float32[:], int64[:], int64[:], int64), parallel=True)
def random_projection(
    data: np.ndarray, indices: np.ndarray, indptr: np.ndarray, d: int
):
    """
    Randomly project a sparse matrix (standard CSR representation) to a dense vectors
    by effectively multiplying by a random matrix whose elements are
    :math:`\pm 1 / \sqrt{d}`. The projection matrix is never stored in memory. Instead
    the elements are generated as needed using splitmix64().
    **Note** Embedding dimension, `d`, is rounded to a multiple of 64 since we generate
    random bits in batches of 64 in order to utilize bit manipulation to speed things
    up.

    The column indices for row i are stored in indices[indptr[i]:indptr[i+1]] and the
    corresponding values are stored in data[indptr[i]:indptr[i+1]].

    Parameters
    ----------
    data : np.ndarray, shape=(nnz,), dtype=float32
        The values of the nonzero elements in the sparse csr_matrix where nnz = number
        of nonzero elements
    indices : np.ndarray, shape=(nnz,), dtype=int64
        The column indices of the nonzero elements in the sparse csr_matrix
    indptr : np.ndarray, shape=(n_rows+1,), dtype=int64
        Pointers into `data` and `indices` to indicate where the rows start and stop.
        If you have just a single record, then indtpr=[0, len(data)]
    d : int
        Embedding dimension of dense vectors.

    Returns
    -------
    X : np.ndarray, shape=(n_rows, d), dtype=float32
        Dense 2-d array containing the randomly projected dense vectors for each row of
        the input sparse matrix.
    """
    assert data.shape == indices.shape, "Shape of data and indices do not match"
    assert data.ndim == 1, "data and indices must be 1-d"
    assert indptr.ndim == 1, "indptr must be 1-d"

    d = max(64, d // 64 * 64)
    # This preserves the magnitude of the vector
    scale = float32(1.0 / np.sqrt(d))
    n_rows = indptr.shape[0] - 1
    X = np.zeros((n_rows, d), float32)

    # Iterate through the rows in parallel
    for row_id in prange(n_rows):
        # Get the slices that correspond to just row_id
        index = indices[indptr[row_id] : indptr[row_id + 1]]
        value = data[indptr[row_id] : indptr[row_id + 1]]
        # Accumulate the matrix-vector project one column at a time
        for i in range(index.size):
            # Incrementally generate a column of the random matrix, 64 bits at a time
            # Multiply it by the corresponding element of the sparse vector
            # Accumulate the result in the dense projection vector
            n_chunks = d // 64
            for chunk in range(n_chunks):
                bits = splitmix64(index[i] * n_chunks + chunk)
                for bitpos in range(64):
                    sign = ((bits >> bitpos) & 1) * 2 - 1
                    X[row_id, chunk * 64 + bitpos] += sign * scale * value[i]

    return X


def distributional_johnson_lindenstrauss_optimal_delta(
    sparse_dim: int, n_components: int, eps: float
) -> float:
    """
    Algorithm to find the optimal failure rate, delta, for a given ``eps`` (error rate),
    ``sparse_dim``, and ``n_components`` (embedding dimension) for the Distributional
    Johnson-Lindenstrauss Lemma which is a different formulation of the problem.
    Taken from

    M. Skorski. *Johnson-Lindenstrauss Transforms with Best Confidence*,
    Proceedings of Machine Learning Research **134**, 1 (2021)

    which can be found at http://proceedings.mlr.press/v134/skorski21a/skorski21a.pdf

    If :math:`\mathbf{A}` is the random projection matrix, then :math:`\delta` is the
    probability of exceeding error limits given by

    .. math::

        \delta = \mathbb{P} \lbrack \\vert \Vert \mathbf{A}x \Vert^2_2 -
        \Vert x \Vert^2_2 \\vert > \epsilon \Vert x \Vert^2_2 {\\rbrack}


    Parameters
    ----------
    sparse_dim : int
        Original dimension of the data
    n_components : int
        Embedding dimension
    eps : float
        Error rate

    Returns
    -------
    delta : float
        The best probability of failure that you can expect
    """
    a = n_components / 2.0
    b = (sparse_dim - n_components) / 2.0
    z0 = a / (a + b)
    dist = beta(a, b)
    fun = lambda z: -dist.cdf((1.0 + eps) * z) + dist.cdf((1.0 - eps) * z)
    betainc_jac = lambda z: dist.pdf(z)
    jac = lambda z: -(1.0 + eps) * betainc_jac((1.0 + eps) * z) + (
        1.0 - eps
    ) * betainc_jac((1.0 - eps) * z)
    out = minimize(fun, x0=z0, jac=jac, method="Newton-CG")
    scale, delta = out.x, 1.0 + out.fun

    # If you do sparse_dim = 2**31 - 1 this returns a numpy array for some reason
    # If you do sparse_dim <= 2**30 it gives you a float
    if isinstance(delta, np.ndarray):
        return delta[0]
    else:
        return delta


def random_sparse_vectors(
    n_samples: int,
    *,
    sparse_dim: int = 2**31 - 1,
    min_n_features: int = 20,
    max_n_features: int = 100,
    normalize: bool = False,
    rng: np.random.Generator = None,
) -> csr_matrix:
    """
    Randomly generate sparse vectors for testing

    Parameters
    ----------
    n_samples : int
        Number of sparse vectors to create
    sparse_dim : int, default = 2\*\*31 - 1
        Dimensionality of the sparse vectors.
    min_n_features : int, default = 10
        Minimum number of features any given sparse vector may have
    max_n_features : int, default = 100
        Maximum number of features any given sparse vector may have
    normalize : bool, default = False
        If true, normalizes the sparse vectors to all have unit length.
    rng : np.random.Generator, default = None
        A numpy random generator. If None, then one is created

    Returns
    -------
    csr_matrix
        Shape is (n_samples, sparse_dim)
    """
    if rng is None:
        rng = np.random.default_rng()

    if sparse_dim >= 2**31:
        raise ValueError(
            f"{sparse_dim=:,} must be below 2**31 due to csr_matrix limits"
        )
    n_features = rng.integers(min_n_features, max_n_features, n_samples)
    data = rng.uniform(1.0, 20.0, np.sum(n_features)).astype(np.float32)
    data = data * rng.choice([-1, 1], np.sum(n_features)).astype(np.float32)
    indices = rng.integers(0, sparse_dim, np.sum(n_features)).astype(np.int32)
    indptr = np.concatenate((np.zeros(1), np.cumsum(n_features))).astype(np.int32)

    # Maybe normalize, but make sure we don't have repeat indices in a given vector
    for i in range(n_samples):
        start_idx = indptr[i]
        end_idx = indptr[i + 1]
        vec_indices = indices[start_idx:end_idx]
        while len(vec_indices) != len(set(vec_indices)):
            indices[start_idx:end_idx] = rng.integers(0, sparse_dim, n_features[i])
            vec_indices = indices[start_idx:end_idx]
        if normalize:
            data[start_idx:end_idx] = data[start_idx:end_idx] / np.linalg.norm(
                data[start_idx:end_idx]
            )

    return csr_matrix((data, indices, indptr), shape=(n_samples, sparse_dim))


class JustInTimeRandomProjection(BaseRandomProjection):
    """
    Reduce the dimensionality of sparse vectors using a dense random projection matrix
    in a memory-efficient way.

    This is an alternative to scikit-learn.random_projection.SparseRandomProjection.
    This implementation provides higher quality embedding and is constant in the amount
    of RAM needed, irrespective of the sparse starting dimension. It achieves this by
    not storing the projection matrix in memory, but instead generates only the needed
    columns just-in-time for any given input vector. This makes the same as a
    SparseRandomProjection when the density = 1.0.

    The elements of the projection matrix, :math:`\pm 1 / \sqrt{n\_components}`, are
    generated as needed using a hashing function to provide a random choice of the
    sign. This uses numba functions to enable faster computation and is parallelized
    across the available cpus.

    This allows you to utilize sparse starting dimension of 2\*\*31 - 1 without any
    issues. For really values of ``n_components`` transforming will also be faster
    than SparseRandomProjection.

    Parameters
    ----------
    n_components : int or 'auto', default = 'auto'
        Dimensionality of the target projection space. If not a multiple of 64, then
        it will be rounded down to the nearest multiple of 64.

        If 'auto', then it will pick a dimensionality that satisfies the
        Johnson-Lindenstrauss Lemma based upon the ``eps`` parameter. The
        dimensionality will then be rounded up to the nearest multiple of 64
        to ensure you satisfy the conditions in the lemma.

        **NOTE** This can yield a very conservative estimate of the required
        dimensionality.
    eps : float, default = 0.1
        Parameter to control the quality of the embedding according to the
        Johnson-Lindenstrauss Lemma. This is used if ``n_components`` is 'auto'.
        This represents the error rate between distances in the sparse dimension
        and the resulting lower embedding dimension. Smaller values of ``eps``
        give larger values of ``n_components``.

    Attributes
    ----------
    n_components_ : int
        Dimensionality of the embedding dimension. It will be a multiple of 64.

    Examples
    --------

    ::

        import numpy as np
        from pocket_dimension.random_projection import (
            JustInTimeRandomProjection,
            random_sparse_vectors,
        )

        n_components = 256
        sparse_dim = 1_073_741_824
        n_samples = 100_000
        X = random_sparse_vectors(n_samples, sparse_dim=sparse_dim, normalize=True)

        transformer = JustInTimeRandomProjection(n_components)
        # No need to fit if provide a value for n_components at initialization
        X_new = transformer.transform(X)
        X_new.shape
        # (25, 64)

    """

    def __init__(self, n_components="auto", *, eps=0.1):
        super().__init__(
            n_components=n_components,
            eps=eps,
            compute_inverse_components=False,
            random_state=None,
        )
        if isinstance(self.n_components, int):
            self.n_components = max(64, (self.n_components // 64) * 64)
            self.n_components_ = self.n_components
            if self.n_components != n_components:
                warnings.warn(
                    "The number of components has been round down to the nearest "
                    + f"multiple of 64. It's been set to {self.n_components}"
                )
        elif self.n_components != "auto":
            raise ValueError(
                f"{n_components=:}. It must either be a positive integer or 'auto'"
            )

    def fit(self, X, y=None):
        """
        This is essential a no-op function since there is no need to generate a random
        projection matrix. Instead, this is validating input data and if
        ``n_components`` is 'auto', then it sets it according to the
        Johnson-Lindenstrauss lemma limits based upon X.shape[1] and ``eps``.

        Parameters
        ----------
        X : csr_matrix
            Only the shape is used to find the optimal value of ``n_components`` to
            satisfy Johnson-Lindenstrauss lemma theoretical guarantees.
        y : Ignored

        Returns
        -------
        self : object
            JustInTimeRandomProjection class instance
        """
        X = self._validate_data(X, accept_sparse="csr", dtype=[np.float32, np.float64])
        n_samples, n_features = X.shape

        if self.n_components == "auto":
            self.n_components_ = johnson_lindenstrauss_min_dim(
                n_samples=n_samples, eps=self.eps
            )
            self.n_components_ = max(64, (1 + self.n_components_ // 64) * 64)

            if self.n_components_ > n_features:
                raise ValueError(
                    f"eps={self.eps:f} and {n_samples=:} lead to a target dimension "
                    + f"{self.n_components_} which is larger than the original space "
                    + f"with {n_features=:}"
                )
        elif self.n_components > n_features:
            warnings.warn(
                "The number of components is higher than the number of features: "
                + f"n_features < n_components ({n_features} < {self.n_components}. "
                + "The dimensionality of the problem will not be reduced.",
                DataDimensionalityWarning,
            )

        return self

    def transform(self, X: csr_matrix) -> np.ndarray:
        """
        Project the data using the just-in-time generated dense random projection
        matrix.

        Parameters
        ----------
        X : csr_matrix
            The input data to project into a smaller dimensional space. Shape is
            (n_samples, sparse_dim)

        Returns
        -------
        np.ndarray
            Projected data of shape (n_samples, n_components)
        """
        check_is_fitted(self)
        X = self._validate_data(
            X, accept_sparse="csr", reset=False, dtype=[np.float32, np.float64]
        )
        if not isinstance(X, csr_matrix):
            raise TypeError(f"X must be a sparse csr_matrix. You gave {type(X)}")

        input_dtype = X.data.dtype

        return random_projection(
            X.data.astype(np.float32),
            X.indices.astype(np.int64),
            X.indptr.astype(np.int64),
            self.n_components_,
        ).astype(input_dtype)

    def _make_random_matrix(self, n_components, n_features):
        """
        This is an abstract method in the base class, so it must be overwritten, but
        it is not need, so returning an empty numpy array.

        Parameters
        ----------
        n_components : int
            The embedding dimension
        n_features : int
            The sparse starting dimension

        Returns
        -------
        np.ndarray
            Empty array to conform to BaseRandomProjection, the parent class
        """
        return np.array([])



"""
Pocket Dimension provides a memory-efficient, dense, random projection of sparse vectors
and then applies this to Term Frequency (TF) and Term Frequency, Inverse Document
Frequency (TFIDF) data.

Copyright (C) 2022 Matthew Hendrey & Brendan Murphy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from numba import njit, float32
import numpy as np
import logging
from pathlib import Path
from pybloomfilter import BloomFilter
from sketchnu.countmin import load as load_cms, CountMin
from sklearn.feature_extraction import FeatureHasher
from typing import Dict, Iterable, List, Tuple, Union

from .random_projection import JustInTimeRandomProjection

logger = logging.getLogger("pocket_dimension")


@njit(float32(float32, float32))
def numba_idf(doc_freq, n_records):
    """
    Numba function to calculate the idf value for a ``doc_freq``

    Parameters
    ----------
    doc_freq : float32
        Number of records that contain this particular feature.
    n_records : float32
        Total number of records in the data set.

    Returns
    -------
    float32
    """
    idf = np.log10(max(float32(1.0), n_records / (float32(1.0) + doc_freq)))
    return idf


class TFVectorizer:
    """
    Calculate a Term-Frequency vector representation by first using sklearn's
    FeatureHasher to create a high-dimensional, sparse vector representation.
    Then apply the JustInTimeRandomProjection to return a lower-dimensional
    dense vector representation. The resulting vectors will be normalized to unit
    length.

    Parameters
    ----------
    d : int
        Dimension of the dense vector output. Rounded down to nearest multiple of 64 if
        not already
    cms_file : str | Path, optional
        Filename of a saved count-min sketch to use for filtering out features
        whose document frequency falls outside of [minDF, maxDF]. Default is None
    hash_dim : int, optional
        Number of dimensions that features get hashed to for the sparse vector
        representation. Should have `hash_dim >> d`. Default is the largest
        possible 2\*\*31 - 1 (~2 billion) for sklearn's FeatureHasher
    minDF : int, optional
        Minimum document frequency (number of records with this feature) a feature
        must have to be included in the embedding. Default is 1
    maxDF : int, optional
        Maximum document frequency (number of records with this feature) a feature
        can have to be included in the embedding. Default is 2\*\*32 - 1.
    temperature : float, optional
        Option to reshape the term frequency vector by raising each count to
        (1/temperature). Using a value above 1.0 flattens the values relative to
        each other. Using a value below 1.0 sharpens the contrast of values
        relative to each other.
    min_n_features : int
        Must have at least this many features in order to create a vector. Useful as a
        data quality check. Default is 1
    min_n_observations : int
        Must have at least his many observations in order to create a vector. Useful as
        a data quality check. This is the sum of counts over the different features for
        a given item. Default is 1
    filter : pybloomfilter3, option
        Bloom filter to use for filtering features before creating sparse vectors.
        Default is None
    filter_out : bool, optional
        If True, then remove features that are in the filter before making the
        sparse vector. If False, then only use features that are in the filter in
        the sparse vector. Not used if `filter` is None. Default is True.

    Attributes
    ----------
    d : int
        Embedding dimension of the dense vector. Must be multiple of 64
    hash_dim : int
        Dimension of the sparse vector. Must be <= 2\*\*31-1
    hasher : sklearn.feature_extraction.FeatureHasher
        Hashes input features (bytes) to integer representing corresponding dimension
    jit_rp : JustInTimeRandomProjection
        Efficiently projects high-dimensional sparse vectors down to lower dimensional
        dense vectors
    temperature : float
        Exponential scale raw counts by 1 / temperature when making TF vector
    min_n_features : int
        Must have at least this many features in order to create a vector
    min_n_observations : int
        Must have at least his many observations in order to create a vector
    cms : sketchnu.CountMinLinear | sketchnu.CountMinLog16 | sketchnu.CountMinLog8
        A count-min sketch that stores document frequency info
    minDF : int
        Minimum document frequency a feature must have to be incuded in the vector
    maxDF : int
        Maximum document frequency a feature can have to be included in the vector
    filter : pybloomfilter.BloomFilter
        BloomFilter containing features to be filtered
    filter_out : bool
        If True, then any feature found in the `filter` will be excluded from the
        vector. If False, then only features found in the `filter` are included in
        vector.
    """

    def __init__(
        self,
        d: int,
        *,
        cms_file: Union[str, Path] = None,
        hash_dim: int = 2**31 - 1,
        minDF: int = 1,
        maxDF: int = 2**32 - 1,
        temperature: float = 1.0,
        min_n_features: int = 1,
        min_n_observations: int = 1,
        filter: str = None,
        filter_out: bool = True,
    ):
        """
        Initialize the a Term-Frequency vectorizer. All optional parameters must be
        passed as keyword arguments. That is, everything but the embedding dimension
        `d`.

        Parameters
        ----------
        d : int
            Dimension of the dense vector output. Gets converted to multiple of 64 if
            not already
        cms_file : str | Path, optional
            Filename of a saved count-min sketch to use for filtering out features
            whose document frequency falls outside of [minDF, maxDF]. Default is None
        hash_dim : int, optional
            Number of dimensions that features get hashed to for the sparse vector
            representation. Should have `hash_dim >> d`. Default is the largest
            possible 2\*\*31 - 1 (~2 billion) for sklearn's FeatureHasher
        minDF : int, optional
            Minimum document frequency (number of records with this feature) a feature
            must have to be included in the embedding. Default is 1
        maxDF : int, optional
            Maximum document frequency (number of records with this feature) a feature
            can have to be included in the embedding. Default is 2\*\*32 - 1.
        temperature : float, optional
            Option to reshape the term frequency vector by raising each count to
            (1/temperature). Using a value above 1.0 flattens the values relative to
            each other. Using a value below 1.0 sharpens the contrast of values
            relative to each other.
        min_n_features : int, optional
            Must have at least this many features in order to create a vector. Useful
            as a data quality check. Default is 1
        min_n_observations : int, optional
            Must have at least his many observations in order to create a vector.
            Useful as a data quality check. This is the sum of counts over the
            different features for a given item. Default is 1
        filter : pybloomfilter3, option
            Bloom filter to use for filtering features before creating sparse vectors.
            Default is None
        filter_out : bool, optional
            If True, then remove features that are in the filter before making the
            sparse vector. If False, then only use features that are in the filter in
            the sparse vector. Not used if `filter` is None. Default is True.
        """
        if d % 64 != 0:
            d = 64 * max(1, int(d // 64))
            logger.warning(f"d is not a multiple of 64. Changing it to {d}")
        assert d < hash_dim, f"{d=:} must be less than {hash_dim=:}"
        assert temperature > 0.0, f"{temperature=:} must be non-negative value"
        assert min_n_features > 0, f"{min_n_features=:} must be positive integer"
        assert (
            min_n_observations > 0
        ), f"{min_n_observations=:} must be positive integer"
        assert minDF <= maxDF, f"{minDF} must be <= {maxDF}"

        self.d = d
        self.hash_dim = hash_dim
        self.minDF = minDF
        self.maxDF = maxDF
        self.temperature = temperature
        self.min_n_features = min_n_features
        self.min_n_observations = min_n_observations
        self.filter_out = filter_out
        self.one_over_temp = 1.0 / temperature
        self.hasher = FeatureHasher(
            hash_dim, input_type="pair", alternate_sign=True, dtype="float32"
        )
        self.jit_rp = JustInTimeRandomProjection(n_components=self.d)

        if filter is None:
            self.filter = []
        elif isinstance(filter, str):
            self.filter = BloomFilter.open(filter, "r")
        else:
            raise ValueError(f"filter must be str | None. You gave {type(filter)}")

        if cms_file is None:
            self.cms = CountMin(width=1, depth=1, cms_type="linear")
            self.cms.add(b"anything")
        elif isinstance(cms_file, (str, Path)):
            self.cms = load_cms(cms_file)

    def __call__(self, records: Iterable[Dict]) -> Tuple[np.ndarray, np.ndarray]:
        """
        Transform the records into sparse vectors with the FeatureHasher and then apply
        the random projection. The vectors are then normalized to unit length.

        Parameters
        ----------
        records: Iterable[Dict]
            Iterable of records where each record has
            {"id": str, "features": List[bytes], "counts": List[int]}

        Returns
        -------
        X : np.ndarray, shape = [n_records, d]
            Resulting dense vector representation of the records. Each vector is
            normalized to unit length.
        ids : np.ndarray, shape = [n_records,]
            Array listing the record's id in order to track which row in X corresponds
            with which id.
        """
        if not isinstance(records, Iterable):
            raise TypeError(
                f"records, {type(records)}, must be an instance of Iterable"
            )
        ids = []
        F = self.yield_record(records, ids)
        H = self.hasher.transform(F)
        X = self.jit_rp.transform(H)

        norms = np.linalg.norm(X, axis=1, keepdims=True)
        keep_nonzero = (norms > 0.0).reshape(-1)
        # Remove records that are all zeros (they had no features after filtering)
        X = X[keep_nonzero, :] / norms[keep_nonzero, :]
        ids = np.array(ids)[keep_nonzero]
        # X = X / np.linalg.norm(X, axis=1, keepdims=True)

        return X, ids

    def _idf(self, doc_freq: float) -> float:
        """
        Return the inverse document frequency given ``doc_freq``

        **NOTE** For TFVectorizer, this simply returns 1.0. Included here for
        compatability with TFIDFVectorizer

        Parameters
        ----------
        doc_freq : float
            A document frequency

        Returns
        -------
        idf : float
            The inverse document frequency
        """
        return 1.0

    def filter_reweight_features(
        self, features: List[bytes], counts: List[int]
    ) -> List[Tuple[bytes, float]]:
        """
        Filter the features accordingly and reweight the associated counts. The counts
        are scaled by raising them to :math:`1/temperature` and then multiplied by the
        ``idf`` value. For TFVectorizer :math:`idf=1.0`.

        Features will be filtered out if their document frequency is not within
        [``minDF``, ``maxDF``] and if a ``filter`` is provided during initialization.
        Only those features that pass the filters will contribute to the final vector.

        After filtering, a final check is done to ensure that there are at least
        ``min_n_features`` features remaining and that there are at least
        ``min_n_observations`` observations of those features. Otherwise an empty list
        is returned.

        Parameters
        ----------
        features : List[bytes]
            List of the individual features. These will be hashed by the FeatureHasher
            to turn into a sparse vector representation
        counts : List[int]
            Number of times that feature is present in a given record.

        Returns
        ------
        List[Tuple[bytes, float]]
        """
        values = np.array(counts) ** self.one_over_temp

        features_values = []
        n_features = 0
        n_observations = 0
        for i in range(len(features)):
            f = features[i]
            v = values[i]
            doc_freq = self.cms[f]
            if ((f in self.filter) != self.filter_out) and (
                self.minDF <= doc_freq and doc_freq <= self.maxDF
            ):
                features_values.append((f, v * self._idf(doc_freq)))
                n_features += 1
                n_observations += counts[i]

        if (
            n_features >= self.min_n_features
            and n_observations >= self.min_n_observations
        ):
            return features_values
        else:
            return []

    def yield_record(self, records: Iterable[Dict], ids: List):
        """
        Yields a list of an individual record's (features, values) where the values are
        the counts after reweighting them by ``temperature`` and/or ``idf``.

        Parameters
        ----------
        records : Iterable[Dict]
            An iterable of records where each records is a dict of
            {"id": str, "features": List[bytes], "counts": List[int]}
        ids : List
            Pass in an empty list which will be returned with corresponding ids from
            the `records` passed in

        Yields
        ------
        List[Tuple[bytes, float]]
            Of the appropriate individual (feature, value) for a given record
        """
        for rec in records:
            ids.append(rec["id"])
            yield self.filter_reweight_features(rec["features"], rec["counts"])


class TFIDFVectorizer(TFVectorizer):
    """
    Randomly project records by first applying a high-dimensional feature hasher
    to create a sparse vector representation of the tf-idf and then applying the
    random projection to a dense embedding dimension.

    .. math::

        tfidf = c^{1/temp} \log{10}(n\_records / (doc\_freq + 1))

    """

    def __init__(
        self,
        d: int,
        cms_file: Union[str, Path],
        *,
        hash_dim: int = 2**31 - 1,
        minDF: int = 1,
        maxDF: int = 2**32 - 1,
        temperature: float = 1.0,
        min_n_features: int = 1,
        min_n_observations: int = 1,
        filter: str = None,
        filter_out: bool = True,
    ):
        """
        Initialize a Term-Frequency, Inverse Document-Frequency vectorizer. All
        optional parameters must be passed as keyword arguments. That is, everything
        but `d` and `cms_file`.

        Parameters
        ----------
        d : int
            Dimension of the dense vector output. Gets converted to multiple of 64 if
            not already
        cms_file : str | Path
            Filename of a saved count-min sketch to use for filtering out features
            whose document frequency falls outside of [minDF, maxDF]. Default is None
        hash_dim : int, optional
            Number of dimensions that features get hashed to for the sparse vector
            representation. Should have `hash_dim >> d`. Default is the largest
            possible 2\*\*31 - 1 (~2 billion) for sklearn's FeatureHasher
        minDF : int, optional
            Minimum document frequency (number of records with this feature) a feature
            must have to be included in the embedding. Default is 1
        maxDF : int, optional
            Maximum document frequency (number of records with this feature) a feature
            can have to be included in the embedding. Default is 2\*\*32 - 1.
        temperature : float, optional
            Option to reshape the term frequency vector by raising each count to
            (1/temperature). Using a value above 1.0 flattens the values relative to
            each other. Using a value below 1.0 sharpens the contrast of values
            relative to each other.
        min_n_features : int, optional
            Must have at least this many features in order to create a vector. Useful
            as a data quality check. Default is 1
        min_n_observations : int, optional
            Must have at least his many observations in order to create a vector.
            Useful as a data quality check. This is the sum of counts over the
            different features for a given item. Default is 1
        filter : pybloomfilter3, option
            Bloom filter to use for filtering features before creating sparse vectors.
            Default is None
        filter_out : bool, optional
            If True, then remove features that are in the filter before making the
            sparse vector. If False, then only use features that are in the filter in
            the sparse vector. Not used if `filter` is None. Default is True.
        """
        super(TFIDFVectorizer, self).__init__(
            d,
            cms_file=cms_file,
            hash_dim=hash_dim,
            minDF=minDF,
            maxDF=maxDF,
            temperature=temperature,
            min_n_features=min_n_features,
            min_n_observations=min_n_observations,
            filter=filter,
            filter_out=filter_out,
        )

    def _idf(self, doc_freq: float) -> float:
        """
        Return the inverse document frequency given ``doc_freq``. To speed things up
        this calls a numba function.

        :math:`idf = \log_{10}(\max(1.0, N / (doc\_freq + 1.0)))`

        where :math:`N` is the total number of documents (or records in this case). The
        max is used because we are using a count-min sketch to estimate the document
        frequencies which may result in the denominator be bigger than N.

        Parameters
        ----------
        doc_freq : float
            A document frequency

        Returns
        -------
        idf : float
            The inverse document frequency
        """
        return numba_idf(doc_freq, self.cms.n_records())



"""
Pocket Dimension provides a memory-efficient, dense, random projection of sparse vectors
and then applies this to Term Frequency (TF) and Term Frequency, Inverse Document
Frequency (TFIDF) data.

Copyright (C) 2022 Matthew Hendrey & Brendan Murphy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
import numpy as np
from pybloomfilter import BloomFilter
from pytest import approx
from sketchnu.countmin import CountMin
from scipy.sparse import csr_matrix
from sklearn.random_projection import johnson_lindenstrauss_min_dim
from typing import Tuple


from pocket_dimension.random_projection import (
    random_sparse_vectors,
    random_projection,
    distributional_johnson_lindenstrauss_optimal_delta,
)
from pocket_dimension.vectorizer import (
    numba_idf,
    TFVectorizer,
    TFIDFVectorizer,
)


def sparse_distance_squared(S1: csr_matrix, S2: csr_matrix) -> float:
    """
    Calculate the distance between two sparse vectors
    """
    total = 0.0
    for idx, value in zip(S1.indices, S1.data):
        if idx in S2.indices:
            total += (value - S2.data[idx]) ** 2
        else:
            total += value**2
    for idx, value in zip(S2.indices, S2.data):
        if idx in S1.indices:
            total += 0.0  # already accounted for above
        else:
            total += value**2
    return total


def test_johnson_lindenstrauss(
    n: int = 40,
    min_n_features: int = 5,
    max_n_features: int = 100,
    eps: float = 0.05,
):
    """
    Test the the Johnson-Lindenstrauss Lemma holds between the sparse vectors and the
    randomly projected dense vectors. This uses the results found in

    https://cs.stanford.edu/people/mmahoney/cs369m/Lectures/lecture1.pdf

    This checkes that the ratio of the distance between two vectors in the embedding
    to the distance between the original sparse vectors is between (1-eps) and (1+eps).
    The embedding dimension is determined by the requested error rate, eps.

    d = 24 \* log(n) / (3 \* eps\*\*2 - 2 \* eps \*\* 3)

    where n is the number of vectors.  This does the full n(n-1)/2 comparisions.

    Parameters
    ----------
    n : int, optional
        Number of sparse vectors to randomly generate. Default is 40
    min_n_features : int, optional
        Minimum number of features any given sparse vector may have. Default is 5
    max_n_features : int, optional
        Maximum number of features any given sparse vector may have. Default is 100
    eps : float, optional
        Desired error rate. Default is 0.05
    """
    # Johnson-Lindenstrauss Lemma says that the embedding dimension must be
    # at least bigger than this value.
    d = johnson_lindenstrauss_min_dim(n, eps=eps)
    # random_project needs multiple of 64. So make sure we satisfy minimum
    # requirement (hence the + 1)
    d = int(d // 64 + 1) * 64

    S = random_sparse_vectors(
        n, min_n_features=min_n_features, max_n_features=max_n_features
    )
    X = random_projection(
        S.data.astype(np.float32),
        S.indices.astype(np.int64),
        S.indptr.astype(np.int64),
        d,
    )

    for i in range(0, X.shape[0] - 1):
        for j in range(i + 1, X.shape[0]):
            sparse_distance2 = sparse_distance_squared(S[i], S[j])
            dense_distance2 = np.sum(np.square(X[i] - X[j]))
            ratio = dense_distance2 / sparse_distance2
            assert (1 - eps) <= ratio and ratio <= (
                1 + eps
            ), f"{ratio=:.4f} is outside of [{1-eps},{1+eps}]"


def test_distributional_johnson_lindenstrauss(
    n: int = 1000,
    min_n_features: int = 5,
    max_n_features: int = 100,
    eps: float = 0.1,
    delta_pad: float = 0.01,
):
    """
    Test the statistical guarantees of the Distributional Johnson-Lindenstrauss Lemma
    using a method for finding the best possible delta (failure rate) for a given
    epsilon (error rate) of the difference of the L2 norm between vectors in the
    original dimension (x) and the lower embedding dimension (Ax) described in

    M. Skorski. *Johnson-Lindenstrauss Transforms with Best Confidence*,
    Proceedings of Machine Learning Research **134**, 1 (2021)
    http://proceedings.mlr.press/v134/skorski21a/skorski21a.pdf

    This paper provides the optimal possible error probability

    delta(sparse_dim, embed_dim, eps) =
            P[abs(\|Ax\|\*\*2 - \|x\|\*\*2) > eps * \|x\|\*\*2]

    given the best possible matrix A and the worst possible data x.

    Parameters
    ----------
    n : int, optional
        Number of data points. Default is 1,000
    min_n_features : int, optional
        Minimum number of features any given vector may have. Default is 5
    max_n_features : int, optional
        Maximum number of features any given vector may have. Default is 100
    eps : float, optional
        Error rate. Default is 0.1
    delta_pad : float, optional
        Amount by which we pad the calculated best delta value. This is done to reduce
        the probability of failure given the vagaries of running statistical tests.
        Default is 0.01
    """
    S = random_sparse_vectors(
        n, min_n_features=min_n_features, max_n_features=max_n_features
    )

    for d in [64, 128, 256, 512, 1024, 2048]:
        X = random_projection(
            S.data.astype(np.float32),
            S.indices.astype(np.int64),
            S.indptr.astype(np.int64),
            d,
        )
        delta = (
            distributional_johnson_lindenstrauss_optimal_delta(2**31 - 1, d, eps)
            + delta_pad
        )
        metric = np.zeros(n)
        for i in range(n):
            x2 = np.sum(np.square(X[i]))
            s2 = np.sum(np.square(S[i].data))
            metric[i] = np.abs(x2 / s2 - 1)
        failure_prob = metric[metric > eps].shape[0] / metric.shape[0]
        assert failure_prob <= delta, f"{failure_prob=:.4f} exceeded {delta=:.4f}"


def test_tf(d: int = 64):
    """
    Basic test that things look right.

    Parameters
    ----------
    d : int, optional
        Embedding dimension. Default is 64
    """
    records = [
        {"id": "one", "features": [b"a", b"b", b"c"], "counts": [1, 2, 3]},
        {"id": "two", "features": [b"a", b"b", b"c"], "counts": [1, 2, 3]},
        {"id": "three", "features": [b"1", b"2", b"3"], "counts": [1, 2, 3]},
    ]

    embedder = TFVectorizer(d)
    X, ids = embedder(records)

    assert X.shape == (3, d)
    assert X.dtype == np.float32
    cosine_0_1 = X[0].dot(X[1])
    assert cosine_0_1 == approx(1.0), f"{cosine_0_1=:.5f} should be 1.0"
    assert ids.shape == (3,)
    assert ids[0] == "one"
    assert ids[1] == "two"
    assert ids[2] == "three"
    cosine_0_2 = X[0].dot(X[2])
    assert cosine_0_2 < 0.25, f"{cosine_0_2=:.5f} should be small"


def test_tf_filter(tmp_path, d: int = 64):
    """
    Testing the bloom filter & filter_out parameters for the TFVectorizer
    """
    records = [
        {
            "id": "one",
            "features": [b"a", b"b", b"c", b"d", b"e"],
            "counts": [1, 2, 3, 4, 5],
        },
        {"id": "two", "features": [b"a", b"b", b"c", b"d"], "counts": [1, 2, 3, 4]},
    ]

    """
    Test that with no filtering, the records have cosine != 1.0
    """
    embedder_nofilter = TFVectorizer(d)
    X, _ = embedder_nofilter(records)
    assert X[0].dot(X[1]) != approx(1.0)

    """
    Test that filtering out the "e" features, that the two vectors now have cosine = 1
    """
    # Add the feature b"e" to the filter.
    bloom_file = str(tmp_path / "tf_filter_out.bloom")
    bloom_filter = BloomFilter(200, 0.0001, bloom_file)
    bloom_filter.add(b"e")
    bloom_filter.close()

    embedder = TFVectorizer(d, filter=bloom_file, filter_out=True)
    X, _ = embedder(records)
    assert X[0].dot(X[1]) == approx(1.0), "'e' is filtered out. cosine should = 1.0"

    """
    Test that vectors are built only with features that are in the bloom filter.
    This will use different set of records, but the same bloom filter.
    """
    # It should already have b"e" in it.
    bloom_filter = BloomFilter.open(bloom_file, "rw")
    bloom_filter.update([b"d", b"c", b"b"])
    bloom_filter.close()

    # Only the b"a" & b"f" values are not in the filter and thus should be
    # removed before the embedding
    records = [
        {
            "id": "one",
            "features": [b"a", b"b", b"c", b"d", b"e"],
            "counts": [5, 4, 3, 2, 1],
        },
        {
            "id": "two",
            "features": [b"a", b"b", b"c", b"d", b"e", b"f"],
            "counts": [15, 4, 3, 2, 1, 40],
        },
    ]
    embedder = TFVectorizer(d, filter=bloom_file, filter_out=False)
    X, _ = embedder(records)
    assert X[0].dot(X[1]) == approx(1.0)

    """
    Test that vectors that have all features removed by the filter are not returned
    """
    bloom_filter = BloomFilter.open(bloom_file, "rw")
    bloom_filter.add(b"a")
    # Now only b"f" is not in the filter
    bloom_filter.close()
    embedder = TFVectorizer(d, filter=bloom_file, filter_out=True)
    X, ids = embedder(records)
    assert X.shape[0] == 1, "Only the id='two' should be in the results"
    assert ids[0] == "two", "Only the id='two' should be in the results"


def test_tf_cms(tmp_path, d: int = 64):
    """
    Test that using a count-min sketch for filtering works properly
    """
    records = [
        {
            "id": "one",
            "features": [b"a", b"b", b"c", b"d", b"e"],
            "counts": [1, 2, 3, 4, 5],
        },
        {
            "id": "two",
            "features": [b"a", b"b", b"c", b"d", b"f"],
            "counts": [1, 2, 3, 4, 30],
        },
        {"id": "three", "features": [b"a", b"b", b"c", b"d"], "counts": [1, 2, 3, 4]},
    ]
    cms_file = str(tmp_path / "tf_cms.npz")
    cms = CountMin("linear", width=300)
    cms.update([b"a"] * 5)
    cms.update([b"b"] * 3)  # Matches minDF=3, should be included
    cms.update([b"c"] * 10)
    cms.update([b"d"] * 100)  # Matches maxDF=100, should be included
    cms.update([b"e"] * 500)  # Will exceed maxDF=100
    cms.update([b"f"] * 2)  # Below minDF=3
    cms.save(cms_file)

    """
    Test that with no filtering, the records have cosine != 1.0
    """
    embedder_nofilter = TFVectorizer(d)
    X_nofilter, _ = embedder_nofilter(records)
    for i in range(2):
        assert X_nofilter[i].dot(X_nofilter[2]) != approx(1.0)

    """
    Test that we drop "e" and "f" from records one & two. They should now match
    the unfiltered three record.
    """
    embedder = TFVectorizer(d, cms_file=cms_file, minDF=3, maxDF=100)
    X, _ = embedder(records)
    for i in range(3):
        assert X[i].dot(X_nofilter[2]) == approx(1.0), f"{i} was not filtered"

    """
    Test that we get back no results if min_n_features is not enough
    """
    embedder = TFVectorizer(d, cms_file=cms_file, minDF=3, maxDF=100, min_n_features=5)
    X, _ = embedder(records)
    assert X.shape == (0, d), f"{X.shape} should be (0, {d})"

    """
    Test that we get back results if min_n_features enough
    """
    embedder = TFVectorizer(d, cms_file=cms_file, minDF=3, maxDF=100, min_n_features=4)
    X, _ = embedder(records)
    assert X.shape == (3, d), f"{X.shape} should be (3, {d})"
    for i in range(3):
        assert X[i].dot(X_nofilter[2]) == approx(1.0), f"{i} was not filtered"

    """
    Test that we get back no results if min_n_observations is not met
    """
    embedder = TFVectorizer(
        d,
        cms_file=cms_file,
        minDF=3,
        maxDF=100,
        min_n_features=4,
        min_n_observations=11,
    )
    X, _ = embedder(records)
    assert X.shape == (0, d), f"{X.shape} should be (0, {d})"

    """
    Test that we get back results if min_n_observations is met
    """
    embedder = TFVectorizer(
        d,
        cms_file=cms_file,
        minDF=3,
        maxDF=100,
        min_n_features=4,
        min_n_observations=10,
    )
    X, _ = embedder(records)
    assert X.shape == (3, d), f"{X.shape} should be (3, {d})"
    for i in range(3):
        assert X[i].dot(X_nofilter[2]) == approx(1.0), f"{i} was not filtered"


def test_tf_temp(d: int = 64):
    """
    Test that changing the temperature used during embedding changes the cosine in an
    expected way
    """
    records = [
        {
            "id": "one",
            "features": [b"a", b"b", b"c", b"d", b"e"],
            "counts": [1, 2, 3, 4, 15],
        },
        {
            "id": "two",
            "features": [b"a", b"b", b"c", b"d", b"e"],
            "counts": [15, 4, 3, 2, 1],
        },
        {
            "id": "three",
            "features": [b"a", b"b", b"c", b"d", b"e"],
            "counts": [1, 1, 1, 1, 1],
        },
    ]
    """
    Testing that using infinite temperature squashes all values to 1.0 and thus
    X_0 and X_1 should match X_2.
    """
    embedder_inf = TFVectorizer(d, temperature=np.inf)
    X, _ = embedder_inf(records)
    cosines = [0.0, 0.0, 0.0]
    for i in range(2):
        cosines[i] = X[i].dot(X[2])
        assert cosines[i] == approx(1.0), f"{i} failed to have cosine~1.0"
    cosines[2] = X[0].dot(X[1])

    """
    Test that as we lower the temperature the cosine decreases between the first two
    vectors and the third reference original vector (all ones) and also between just
    the first two vectors
    """
    for temp in [3.0, 2.0, 1.5, 1.0, 0.8]:
        embedder = TFVectorizer(d, temperature=temp)
        Y, _ = embedder(records)
        for i in range(2):
            cosine = Y[i].dot(X[2])
            assert (
                cosine < cosines[i]
            ), f"{temp} {i} has {cosine=:} less than previous {cosines[i]}"
            cosines[i] = cosine
        # Cosine between the first two vectors should also get worse as temp goes down
        cosine = Y[0].dot(Y[1])
        assert cosine < cosines[2]
        cosines[2] = cosine


def test_numba_idf():
    """
    Test that the numba function numba_idf works as expected
    """
    doc_freq = 100
    n_records = 150

    idf_py = np.log10(n_records / (doc_freq + 1))
    idf = numba_idf(doc_freq, n_records)

    assert idf_py == approx(idf)


def test_tfidf(tmp_path, d: int = 64):
    """
    Test that a tfidf vectorization causes one feature's tfidf to be zero and match
    a vector that does not have this feature
    """
    records = [
        {
            "id": "one",
            "features": [b"a", b"b", b"c", b"d", b"e"],
            "counts": [1, 2, 3, 4, 5],
        },
        {
            "id": "two",
            "features": [b"a", b"b", b"c", b"d", b"e"],
            "counts": [1, 2, 3, 4, 5],
        },
        {
            "id": "three",
            "features": [b"a", b"b", b"c", b"d"],
            "counts": [1, 2, 3, 4],
        },
    ]
    cms_file = str(tmp_path / "tf_cms.npz")
    cms = CountMin("linear", width=300)
    cms.n_added_records[1] = 500
    cms.update([b"a"] * 3)
    cms.update([b"b"] * 5)
    cms.update([b"c"] * 10)
    cms.update([b"d"] * 100)
    cms.update([b"e"] * 500)
    cms.save(cms_file)

    """
    Test that with no filtering, the records have cosine != 1.0
    """
    embedder = TFIDFVectorizer(d, cms_file=cms_file)
    X, _ = embedder(records)

    # Now the b"e" should contribute nothing to the vector since idf = 0.0
    assert X[0].dot(X[2]) == approx(1.0)


def test_tf_data_quality(d: int = 64):
    """
    Test that records without minimium data quality (min_n_features &
    min_n_observations) return vectors

    Parameters
    ----------
    d : int, optional
        Embedding dimension. Default is 64
    """
    records = [
        {"id": "one", "features": [b"a", b"b", b"c"], "counts": [6, 6, 6]},
        {"id": "two", "features": [b"a", b"b", b"c"], "counts": [5, 5, 5]},
        {"id": "three", "features": [b"a", b"b", b"c"], "counts": [1, 1, 1]},
        {"id": "four", "features": [b"a", b"b"], "counts": [20, 20]},
        {"id": "five", "features": [b"a", b"b"], "counts": [1, 1]},
    ]

    embedder = TFVectorizer(d, min_n_features=3, min_n_observations=15)
    X, ids = embedder(records)
    assert X.shape == (2, d)
    assert ids[0] == "one"
    assert ids[1] == "two"
    cosine_0_1 = X[0].dot(X[1])
    assert cosine_0_1 == approx(1.0), f"{cosine_0_1=:.5f} should be 1.0"

    embedder = TFVectorizer(d, min_n_features=3, min_n_observations=16)
    X, ids = embedder(records)
    assert X.shape == (1, d)
    assert ids[0] == "one"

    embedder = TFVectorizer(d, min_n_features=2, min_n_observations=3)
    X, ids = embedder(records)
    assert X.shape == (4, d)
    assert ids[0] == "one"
    assert ids[1] == "two"
    assert ids[2] == "three"
    assert ids[3] == "four"
    cosine_0_1 = X[0].dot(X[1])
    assert cosine_0_1 == approx(1.0), f"{cosine_0_1=:.5f} should be 1.0"
    cosine_0_2 = X[0].dot(X[2])
    assert cosine_0_2 == approx(1.0), f"{cosine_0_2=:.5f} should be 1.0"


def test_tfidf_data_quality(tmp_path, d: int = 64):
    """
    Test that records without minimium data quality (min_n_features &
    min_n_observations) return vectors

    Parameters
    ----------
    tmp_path: str
        Local temp directory to save count-min sketch to
    d : int, optional
        Embedding dimension. Default is 64
    """
    cms_file = str(tmp_path / "tfidf_cms.npz")
    cms = CountMin("linear", width=300)
    cms.n_added_records[1] = 500
    cms.update([b"a"] * 3)
    cms.update([b"b"] * 5)
    cms.update([b"c"] * 10)
    cms.update([b"d"] * 100)
    cms.update([b"e"] * 500)
    cms.save(cms_file)

    records = [
        {"id": "one", "features": [b"a", b"b", b"c"], "counts": [6, 6, 6]},
        {"id": "two", "features": [b"a", b"b", b"c"], "counts": [5, 5, 5]},
        {"id": "three", "features": [b"a", b"b", b"c"], "counts": [1, 1, 1]},
        {"id": "four", "features": [b"a", b"b"], "counts": [20, 20]},
        {"id": "five", "features": [b"a", b"b"], "counts": [1, 1]},
    ]

    embedder = TFIDFVectorizer(
        d, cms_file=cms_file, min_n_features=3, min_n_observations=15
    )
    X, ids = embedder(records)
    assert X.shape == (2, d)
    assert ids[0] == "one"
    assert ids[1] == "two"
    cosine_0_1 = X[0].dot(X[1])
    assert cosine_0_1 == approx(1.0), f"{cosine_0_1=:.5f} should be 1.0"

    embedder = TFIDFVectorizer(
        d, cms_file=cms_file, min_n_features=3, min_n_observations=16
    )
    X, ids = embedder(records)
    assert X.shape == (1, d)
    assert ids[0] == "one"

    embedder = TFIDFVectorizer(
        d, cms_file=cms_file, min_n_features=2, min_n_observations=3
    )
    X, ids = embedder(records)
    assert X.shape == (4, d)
    assert ids[0] == "one"
    assert ids[1] == "two"
    assert ids[2] == "three"
    assert ids[3] == "four"
    cosine_0_1 = X[0].dot(X[1])
    assert cosine_0_1 == approx(1.0), f"{cosine_0_1=:.5f} should be 1.0"
    cosine_0_2 = X[0].dot(X[2])
    assert cosine_0_2 == approx(1.0), f"{cosine_0_2=:.5f} should be 1.0"
